<?php
require_once ASD_APPLICATION_PATH . '/models/AbstractModel.php';

class Asd_Model_Amazon extends Asd_Model_AbstractModel
{
    const SEARCH_PAGE_COUNT = 10;
    const SEARCH_PAGE_COUNT_ROOT = 5;

    const THROTTLE_DELAY_SECONDS = 2;
    const THROTTLE_RETRY_COUNT = 5;

    protected $_options = array();

    /**
    * @var Asd_Model_Amazon_Market
    */
    protected $_market;

    protected $_http;

    protected $_credentialFields = array(
        'access_key',
        'secret_access_key',
    );
    protected $_credentialsInitialized = FALSE;

    protected $_apiScheme = 'http';

    protected $_responseGroups = array(
        'Medium',
        'OfferFull',
        'OfferSummary',
        'VariationOffers',
        'Variations',
        'VariationSummary',
        'VariationImages',
        'EditorialReview',
        'Reviews',
    );

    protected $_keywordLog = array();

    public function __construct()
    {
        $this->_market = Asd_Application::getModel('Amazon_Market');
    }

    public function getCategories($parent = NULL)
    {
        $cacheKey = "amazon.categories.{$this->getOption('market')}."
            . (is_null($parent)? '_root' : $parent);
        /**
        * @var Asb_Model_Cache
        */
        $cache = Asd_Application::getModel('Cache');
        if ($result = $cache->getValue($cacheKey)) {
            return $result;
        }

        if (is_null($parent)) {
            // Return root categories
            $result = $this->getMarketConfig()->rootCategories->toArray();
        } else {
            $result = $this->_getSubcategories($parent);
        }

        natcasesort($result);

        $cache->setValue($cacheKey, $result)
            ->save();

        return $result;
    }

    public function getItemData($itemId)
    {
        $params = array(
            'ItemId' => $itemId,
            'IdType' => 'ASIN',
            'ResponseGroup' => implode(',',
                $this->_responseGroups),
        );

        $response = $this->_request('ItemLookup', $params);

        if (isset($response->Items->Request->Errors->Error->Code)
            && "AWS.ECommerceService.ItemNotAccessible" == $response->Items->Request->Errors->Error->Code
        ) {
            return NULL;
        }

        if (!isset($response->Items->Item)) {
            if (isset($response->Items->Request->Errors->Error->Message)) {
                $this->_throwError('Error retrieving item data: '
                    . $response->Items->Request->Errors->Error->Message);
            }
            $this->_throwError('No item data available');
        }

        return $this->_xmlToArray($response->Items->Item);
    }

    //
    // API wrappers
    //

    protected function _getSubcategories($parent)
    {
        $result = array();
        $response = $this->_request('BrowseNodeLookup', array(
            'BrowseNodeId' => $parent,
        ));

        if (isset($response->BrowseNodes->BrowseNode->Children->BrowseNode)) {
            $nodes = $response->BrowseNodes->BrowseNode->Children->BrowseNode;

            // Try to find nodes with IsCategoryRoot
            $traverse = array();
            foreach ($nodes as $key=>$node) {
                if (isset($node->IsCategoryRoot) && $node->IsCategoryRoot) {
                    $traverse[$key] = $node;
                }
            }

            if (!$traverse) {
                foreach ($nodes as $key=>$node ) {
                    $result[(string)$node->BrowseNodeId] = (string)$node->Name;
                }
            } else {
                foreach ($traverse as $key=>$node) {
                    $response = $this->_request('BrowseNodeLookup', array(
                        'BrowseNodeId' => $node->BrowseNodeId)
                    );
                    if (isset($response->BrowseNodes->BrowseNode->Children->BrowseNode) ) {
                        foreach ($response->BrowseNodes->BrowseNode->Children->BrowseNode as $subKey=>$subNode ) {
                            $result[(string)$subNode->BrowseNodeId] = (string)$subNode->Name;
                        }
                    }
                }
            }
        }

        return $result;
    }

    //
    // Item analysis
    //
    public function isItemAvailable(array $item)
    {
        $result = TRUE;
        if (isset($item['OfferSummary'])
            && 0 == $item['OfferSummary']['TotalNew']
            && 0 == $item['OfferSummary']['TotalUsed']
            && 0 == $item['OfferSummary']['TotalCollectible']
            && 0 == $item['OfferSummary']['TotalRefurbished']
            && isset($item['ItemAttributes']['ListPrice'])
        ) {
            $result = FALSE;
            // Check variations
            if (isset($item['Variations']['Item'])) {
                foreach ($item['Variations']['Item'] as $variation) {
                    if (isset($variation['Offers']['Offer'])
                        || isset($variation['Offer'])
                    ) {
                        $result = TRUE;
                        break;
                    }
                }
            }
        }

        return $result;
    }

    //
    // API request
    //

    protected function _request($operation, array $params)
    {
        $this->_initHttp();
        $this->_initCredentials();

        // Main parametrs
        $method = 'GET';
        $host = $this->getMarketConfig()->apiHost;
        $uri = '/onca/xml';

        $iRetry = 0;
        while(true) {
            // additional parameters
            $params['Service'] = 'AWSECommerceService';
            $params['AWSAccessKeyId'] = $this->getOption('access_key');
            $params['AssociateTag'] = $this->hasOption('affiliate_id')
                ? $this->getOption('affiliate_id')
                : '-'; // Dummy value for requests like getting categories
            $params['Timestamp'] = gmdate('Y-m-d\TH:i:s\Z', time() + 24*3600); //workaround for servers with late clock
            $params['Version'] = '2011-08-01';
            $params['Operation'] = $operation;

            // prepare & sign request
            ksort($params);
            $canonicalizedQuery = array();
            foreach ($params as $param=>$value) {
                $param = str_replace('%7E', '~', rawurlencode($param));
                $value = str_replace('%7E', '~', rawurlencode($value));
                $canonicalizedQuery[] = $param . '=' . $value;
            }
            $canonicalizedQuery = implode('&', $canonicalizedQuery);
            $stringToSign = $method . "\n" . $host . "\n" . $uri . "\n" . $canonicalizedQuery;
            $signature = base64_encode(hash_hmac('sha256', $stringToSign, $this->getOption('secret_access_key'), TRUE));
            $signature = str_replace('%7E', '~', rawurlencode($signature));

            $scheme = $this->_apiScheme;
            $url = "{$scheme}://{$host}{$uri}?{$canonicalizedQuery}&Signature={$signature}";

            $this->_http->resetParameters()
                ->setUri($url);

            // do request
            try {
                $response = $this->_http->request($method);
            } catch (AsdZend_Http_Client_Exception $e) {
                $this->_throwError(sprintf(__('Could not connect to Amazon API (%s)', 'amazon-bar-plugin'), $e->getMessage()));
            } catch (Exception $e) {
                $this->_throwError(sprintf(__('Unexpected error while requesting Amazon API (%s)', 'amazon-bar-plugin'), $e->getMessage()));
            }
            if ($response->isError()) {
                if (403 == $response->getStatus()) {
                    $this->_throwError(__('Invalid Amazon credentials', 'amazon-bar-plugin'));
                }

                $xml = @simplexml_load_string($response->getBody());

                if (isset($xml->Error->Code)
                    && 'RequestThrottled' == $xml->Error->Code
                    && $iRetry < self::THROTTLE_RETRY_COUNT
                ) {
                    $iRetry++;
                    sleep(self::THROTTLE_DELAY_SECONDS);
                    continue;
                }

                if (isset($xml->Error->Message)) {
                    $this->_throwError(sprintf(__('Amazon API returned error: %s', 'amazon-bar-plugin'), $xml->Error->Message));
                }

                $this->_throwError(sprintf(__('Amazon API request failed due to HTTP error #%s (%s)', 'amazon-bar-plugin'), $response->getStatus(), $response->getMessage()));
            }

            break;
        }

        $response = @simplexml_load_string($response->getBody());
        if (FALSE === $response) {
            $this->_throwError(__('Error parsing API response', 'amazon-bar-plugin'));
        }

        return $response;
    }

    //
    // Initializers
    //

    protected function _initCredentials()
    {
        if ($this->_credentialsInitialized) {
            return;
        }

        $options = Asd_Application::getModel('Options');
        foreach ($this->_credentialFields as $field) {
            if (!$this->hasOption($field)) {
                $this->setOption($field, $options->getValue($field));
            }
        }

        $this->_credentialsInitialized = TRUE;
    }

    protected function _initHttp()
    {
        if ($this->_http) {
            return;
        }

        // Create HTTP client
        require_once ASD_LIBRARY_PATH . '/AsdZend/Http/Client.php';
        $this->_http = new AsdZend_Http_Client;
        // Set cURL adapter if cURL extension loaded:
        if (extension_loaded('curl')) {
            require_once ASD_LIBRARY_PATH . '/AsdZend/Http/Client/Adapter/Curl.php';
            $this->_http->setAdapter(new AsdZend_Http_Client_Adapter_Curl);
        }

        $this->_http->setConfig($this->getOption('httpClientConfig'));
    }

    protected function _initMarket()
    {
        $this->_market->setCode($this->getOption('market'));
    }

    //
    // Helpers
    //

    public function getMarketConfig()
    {
        return $this->_market->getConfig();
    }

    protected function _throwError($message)
    {
        require_once ASD_APPLICATION_PATH . '/models/Amazon/Exception.php';
        throw new Asd_Model_Amazon_Exception($message);
    }

    protected function _xmlToArray($xml)
    {
        $result = (array) $xml;
        foreach ($result as $key=>$element) {
            if (is_object($element) || is_array($element)) {
                $result[$key] = $this->_xmlToArray($element);
            }
        }

        return $result;
    }

    //
    // Options management
    //

    public function getOption($key)
    {
        if (array_key_exists($key, $this->_options)) {
            return $this->_options[$key];
        }

        $result = Asd_Application::getConfig()->amazon->options->$key;
        if (is_null($result)) {
            throw new Exception("Required option '{$key}' not found");
        }

        if ($result instanceof AsdZend_Config) {
            $result = $result->toArray();
        }

        return $result;
    }

    public function hasOption($key)
    {
        return isset($this->_options[$key]);
    }

    public function setOptions(array $options)
    {
        $this->_options = $options;

        $this->_credentialsInitialized = FALSE;
        if (isset($options['market'])) {
            $this->_initMarket();
        }

        return $this;
    }

    public function setOption($key, $value)
    {
        $this->_options[$key] = $value;

        if ('market' == $key) {
            $this->_initMarket();
        }

        return $this;
    }

    //
    // Getters and setters
    //

    public function getCredentialFields()
    {
        return $this->_credentialFields;
    }

    public function getKeywordLog()
    {
        return $this->_keywordLog;
    }

    public function resetKeywordLog()
    {
        $this->_keywordLog = array();

        return $this;
    }

    public function getHttp()
    {
        $this->_initHttp();

        return $this->_http;
    }
}